from . import TelegramEndpoint
from .commons import DisableNotification, ParseMode, ReplyToMessageId, _ChatOrInlineMessage
from .reply_markup import ReplyMarkup


class FileEndpoint(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    file_type = None
    attributes = []

    def __init__(self, chat_id, file, is_path=True):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }

        if is_path is True:
            self.method = 'POST'
            with open(file, 'r') as f:
                content = f.read()
            self.fields = {
                **self.fields,
                self.file_type: content
            }
        else:
            self.method = 'GET'
            self.args = {
                **self.args,
                self.file_type: file
            }


class SerializableToInputMedia(object):
    serialized_fields = []

    def to_input_media(self):

        serialized = {
            'type': self.file_type,
        }

        # todo
        # upload + attach
        if self.file_type in self.fields:
            raise Exception('Must target updated file')
            pass

        serialized['media'] = self.args[self.file_type]

        attributes = [*self.attributes, *self.serialized_fields]
        for attribute in attributes:
            if attribute in self.args:
                serialized[attribute] = self.args[attribute]

        return serialized


class Caption(object):
    def caption(self, caption):
        self.args = {
            **self.args,
            'caption': caption
        }
        return self


# todo
class Thumb(object):
    pass


class SendPhoto(FileEndpoint, Caption, ParseMode, SerializableToInputMedia):
    file_type = 'photo'
    serialized_fields = ['caption', 'parse_mode']


class SendAudio(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'audio'
    attributes = ['duration', 'performer', 'title']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendDocument(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'document'
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendVideo(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'video'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb', 'supports_streaming']

    def supports_streaming(self):
        self.args = {
            **self.args,
            'supports_streaming': True
        }
        return self


class SendAnimation(FileEndpoint, Caption, ParseMode, Thumb, SerializableToInputMedia):
    file_type = 'animation'
    attributes = ['duration', 'width', 'height']
    serialized_fields = ['caption', 'parse_mode', 'thumb']


class SendVoice(FileEndpoint, Caption, ParseMode):
    file_type = 'animation'
    attributes = ['duration']


class SendVideoNote(FileEndpoint, Thumb):
    file_type = 'video_note'
    attributes = ['duration', 'length']


class SendMediaGroup(TelegramEndpoint, DisableNotification, ReplyToMessageId):
    def __init__(self, chat_id, media_array):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'media': [x.to_input_media() for x in media_array]
        }


class GetFile(TelegramEndpoint):
    # https://api.telegram.org/file/bot<token>/<file_path>
    def __init__(self, file_id):
        super().__init__()
        self.args = {
            'file_id': file_id
        }


class EditMessageCaption(TelegramEndpoint, Caption, ParseMode, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)


class EditMessageMedia(TelegramEndpoint, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id, media):
        super().__init__()
        self.args = {
            'media': media.to_input_media()
        }
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)



from . import TelegramEndpoint
from .commons import DisableNotification, ReplyToMessageId, _ChatOrInlineMessage
from .reply_markup import ReplyMarkup


class SendLocation(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, latitude, longitude):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'latitude': latitude,
            'longitude': longitude
        }

    def live_period(self, live_period):
        self.args = {
            **self.args,
            'live_period': live_period
        }
        return self


class EditMessageLiveLocation(TelegramEndpoint, _ChatOrInlineMessage, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id, latitude, longitude):
        super().__init__()
        self.args = {
            'latitude': latitude,
            'longitude': longitude
        }
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)


class StopMessageLiveLocation(TelegramEndpoint, ReplyMarkup, _ChatOrInlineMessage):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)



from . import TelegramEndpoint


class SerializableToInlineQueryResult(object):
    pass


class AnswerInlineQuery(TelegramEndpoint):
    def __init__(self, inline_query_id, results):
        super().__init__()
        self.args = {
            'inline_query_id': inline_query_id,
            'results': [x.to_inline_query_result() for x in results]
        }

    def cache_time(self, cache_time):
        self.args = {
            **self.args,
            'cache_time': cache_time
        }
        return self

    def is_personal(self, is_personal):
        self.args = {
            **self.args,
            'is_personal': is_personal
        }
        return self

    def next_offset(self, next_offset):
        self.args = {
            **self.args,
            'is_personal': next_offset
        }
        return self

    def switch_pm_text(self, switch_pm_text):
        self.args = {
            **self.args,
            'switch_pm_text': switch_pm_text
        }
        return self

    def switch_pm_parameter(self, switch_pm_parameter):
        self.args = {
            **self.args,
            'switch_pm_parameter': switch_pm_parameter
        }
        return self



import enum
import logging
from asyncio import PriorityQueue, get_event_loop, iscoroutinefunction
from urllib.parse import urlencode

from .requests import get_pool_manager


class ExceptionThrownWhen(enum.Enum):
    checking_update = 10
    executing_handler = 20
    executing_matcher = 30


class Bot(object):
    update_types = [
        ('message', 'message'),
        ('edited_message', 'message'),
        ('channel_post', 'message'),
        ('edited_channel_post', 'message'),
        ('inline_query', 'inline_query'),
        ('chosen_inline_result', 'chosen_inline_result'),
        ('callback_query', 'callback_query'),
        ('shipping_query', 'shipping_query'),
        ('pre_checkout_query', 'pre_checkout_query')
    ]
    discover_type_and_flavor = False

    def __init__(self, token, json_lib=None, max_handlers=1, loop=None):
        """
        The Class(TM).

        :param token: The Telegram-given token
        """

        self.token = token

        if json_lib is not None:
            self.json = json_lib
        else:
            import json
            self.json = json

        self.max_handlers = max_handlers

        if loop is None:
            self.loop = get_event_loop()
        else:
            self.loop = loop

        self.http = get_pool_manager()
        self.logger = logging.getLogger('ubot')
        self.triggers = []
        self.ignore_check_triggers = []
        self.update_queue = PriorityQueue(loop=loop)
        self.base_url = 'https://api.telegram.org:443/bot{}/'.format(token)

    async def api_request(self, method, endpoint, fields=None):
        """
        Wraps the urllib3 request in a more user friendly way, making it async and premitting the base Telegram API url.

        :param method: The HTTP method name (GET, POST, PUT, DELETE, HEAD)
        :param endpoint: A Telegram API endpoint (e.g. sendMessage), omitting
            the first slash
        :param fields: A dict of params to be used in the request (used as
            query params with GET/HEAD/DELETE, used as from parans with
            POST/PUT)
        :param loop: the loop
        :return: The response from the server, in the form of a dict
        """

        url = '{}{}'.format(self.base_url, endpoint)
        res = await self.loop.run_in_executor(None, self.http.request, method, url, fields)

        return res

    async def set_webhook_url(self, url):
        await self.api_request('GET', 'setWebhook?url={}'.format(url))

    async def start(self):  # noqa: C901
        """
        Main loop

        >>> import asyncio
        >>> loop = asyncio.get_event_loop()
        >>> bot = Bot('token')
        >>> loop.run_until_complete(asyncio.wait([
        >>>     loop.create_task(bot.start())
        >>>     # other tasks
        >>> ]))
        >>> loop.run_forever()

        or (if you want to use a custom event loop

        >>> import asyncio
        >>> loop = asyncio.new_event_loop()
        >>> bot = Bot('token', loop=loop)
        >>> loop.run_until_complete(asyncio.wait([
        >>>     loop.create_task(bot.start())
        >>>     # other tasks
        >>> ]))
        >>> loop.run_forever()

        note: single instance, multiple threads unsupported (for now), use an instance of the bot for every thread
        """

        while True:
            update = await self.update_queue.get()

            # get type and flavor if needed or simply leave them to None
            if self.discover_type_and_flavor is True:
                for _type, flavor in self.update_types:
                    if _type in update:
                        update_type, update_flavor = _type, flavor
                        break
                else:
                    update_type, update_flavor = None, None
            else:
                update_type, update_flavor = None, None

            # check the update if the function is implemented and skip if it's not passed
            if hasattr(self, 'check_update'):
                check_update = self.check_update
                try:
                    if iscoroutinefunction(check_update):
                        is_check_passed = await check_update(update, update_type, update_flavor)
                    else:
                        is_check_passed = check_update(update, update_type, update_flavor)
                except Exception as e:
                    when = ExceptionThrownWhen.checking_update
                    await self.handle_exception(update, e, when)
                    continue

                if is_check_passed is True:
                    checked_triggers = [*self.ignore_check_triggers, *self.triggers]
                else:
                    checked_triggers = self.triggers
            else:
                checked_triggers = self.triggers

            # loop through the triggers to find one that matches
            curr_handlers = 0

            for trigger in checked_triggers:
                if curr_handlers > self.max_handlers:
                    break

                match = trigger.match

                try:
                    if iscoroutinefunction(match):
                        is_update_matched = await match(update, update_type, update_flavor)
                    else:
                        is_update_matched = match(update, update_type, update_flavor)
                except Exception as e:
                    when = ExceptionThrownWhen.executing_matcher
                    await self.handle_exception(update, e, when)
                    continue

                if is_update_matched is not True:
                    continue

                curr_handlers += 1

                # if a matching one is found call the corresponding handler
                handle = trigger.handle

                try:
                    if iscoroutinefunction(handle):
                        endpoint = await handle(update, update_type, update_flavor)
                    else:
                        endpoint = handle(update, update_type, update_flavor)
                except Exception as e:
                    when = ExceptionThrownWhen.executing_handler
                    await self.handle_exception(update, e, when)
                    continue

                if endpoint is None:
                    continue

                method, url, args, fields = endpoint

                _url = '{}?{}'.format(url, urlencode(args))
                response = await self.api_request(method, _url, fields)

                # call the callback if existing
                callback = trigger.callback

                if callback is None:
                    continue

                try:
                    if iscoroutinefunction(callback):
                        await callback(response)
                    else:
                        callback(response)
                except Exception as e:
                    when = ExceptionThrownWhen.executing_handler
                    await self.handle_exception(update, e, when)
                    continue

    def push_update(self, update):
        """
        Pushes an update (already json decoded) into the queue.

        :param update: The update to be pushed in the queue
        """

        self.update_queue.put_nowait((update['update_id'], update))

    async def handle_exception(self, update, exception, when):
        """
        Logs the exception and calls the on_exception function if it exists.
        N.B.: this *must* be used only to log *externally* originated exception

        :param update: The update whose processing triggered the exception
        :param exception: The exception object
        :param when: A string containing info about when the Exception was thrown
        """

        _when = str(ExceptionThrownWhen(when).name)
        error = 'Exception {0}\n occurred when {1}\n during handling of' \
                ' update:\n {2}' \
            .format(str(exception), _when, self.json.dumps(update))

        self.logger.exception(error, exc_info=True)

        if hasattr(self, 'on_exception'):
            on_exception = self.on_exception
            if iscoroutinefunction(on_exception):
                await on_exception(update, exception, _when)
            else:
                on_exception(update, exception, _when)

    def trigger(self, requires_type_and_flavor=False, ignore_check=False):
        """
        Decorates a Trigger, inserting it into the bot check list

        :param requires_type_and_flavor: Forces the bot to discover the type and the flavor of an update and to call
            the trigger methods with those as arguments
        :param ignore_check: ignores a possible "check_update" function
        """

        def decorator(trigger):
            if ignore_check is True:
                self.ignore_check_triggers.append(trigger())
            else:
                self.triggers.append(trigger())

            if requires_type_and_flavor is True:
                self.discover_type_and_flavor = True

            return trigger

        return decorator



from . import TelegramEndpoint
from .exceptions import InvalidInlineKeyboardButtonPosition, InvalidInlineKeyboardButtonType


class InlineKeyboard(object):
    def __init__(self):
        self.buttons = []

    def serialize(self):
        return {'inline_keyboard': self.buttons}

    def add_button(self, text, button_type, button_value):
        if button_type not in ['url', 'callback_data', 'switch_inline_query', 'switch_inline_query_current_chat',
                               'callback_game', 'pay']:
            raise InvalidInlineKeyboardButtonType

        if button_type in ['callback_game', 'pay'] and self.buttons:
            raise InvalidInlineKeyboardButtonPosition('You must insert a callback_game/pay button as the first button')

        self.buttons.append({
            'text': text,
            button_type: button_value
        })
        return self


class ReplyKeyboard(object):
    def __init__(self):
        self.buttons = []
        self.options = {}

    def serialize(self):
        return {
            'keyboard': self.buttons,
            **self.options
        }

    def resize_keyboard(self):
        self.options['resize_keyboard'] = True
        return self

    def one_time_keyboard(self):
        self.options['one_time_keyboard'] = True
        return self

    def selective(self):
        self.options['selective'] = True
        return self

    def add_button(self, text, request_contact=False, request_location=False):
        button = {
            'text': text
        }

        if request_contact is True:
            button['request_contact'] = True

        if request_location is True:
            button['request_location'] = True

        self.buttons.append(button)
        return self


class ReplyKeyboardRemove(object):
    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'remove_keyboard': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class ForceReply(object):
    def __init__(self):
        self.options = {}

    def serialize(self):
        return {
            'force_reply': True,
            **self.options
        }

    def selective(self):
        self.options['selective'] = True
        return self


class ReplyMarkup(object):
    def reply_markup(self, reply_markup):
        self.args = {
            **self.args,
            'reply_markup': reply_markup.serialize()
        }
        return self


class AnswerCallbackQuery(TelegramEndpoint):
    def __init__(self, callback_query_id):
        super().__init__()
        self.args = {
            'callback_query_id': callback_query_id
        }

    def text(self, text):
        self.args = {
            **self.args,
            'text': text
        }
        return self

    def show_alert(self):
        self.args = {
            **self.args,
            'show_alert': True
        }
        return self

    def url(self, url):
        self.args = {
            **self.args,
            'url': url
        }
        return self

    def cache_time(self, cache_time):
        self.args = {
            **self.args,
            'cache_time': cache_time
        }



from . import TelegramEndpoint
from .commons import DisableNotification, ReplyToMessageId
from .reply_markup import ReplyMarkup


class SendContact(TelegramEndpoint, DisableNotification, ReplyMarkup, ReplyToMessageId):
    def __init__(self, chat_id, phone_number, first_name):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'phone_number': phone_number,
            'first_name': first_name  # noqa: S001
        }

    def last_name(self, _last_name):
        self.args = {
            **self.args,
            'last_name': _last_name
        }
        return self

    def vcard(self, vcard):
        self.args = {
            **self.args,
            'vcard': vcard
        }
        return self



class OptionalFeatures(object):

    async def check_update(self, update, update_type, update_flavor):
        """
        To be implemented by user, called as validator before an update is handled, if the return value is not True the
        update is not queued

        :param update: The update (a dict) received from the Telegram webhook
        :return: True/Other value
        """
        return True

    async def on_exception(self, update, exception, when):
        """
         To be implemented by user, called after an exception (e.g. to report it on a telegram chat, to send you a mail
         with the error, ecc..)

        :param update: The update (a dict) received from the Telegram webhook
        :param exception: The exception object
        :param when: A string containing info about when the Exception was thrown
        """
        pass



from . import TelegramEndpoint
from .commons import DisableNotification


class ExportChatInviteLink(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class SetChatPhoto(TelegramEndpoint):
    method = 'POST'

    def __init__(self, chat_id, photo):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }

        with open(photo, 'r') as f:
            content = f.read()

        self.fields = {
            **self.args,
            'photo': content
        }


class DeleteChatPhoto(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class SetChatTitle(TelegramEndpoint):
    def __init__(self, chat_id, title):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'title': title
        }


class SetChatDescription(TelegramEndpoint):
    def __init__(self, chat_id, description):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'description': description
        }


class PinChatMessage(TelegramEndpoint, DisableNotification):
    def __init__(self, chat_id, message_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'message_id': message_id
        }


class UnpinChatMessage(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class LeaveChat(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class GetChat(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class GetChatAdministrators(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class GetChatMemberCount(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }


class GetChatMember(TelegramEndpoint):
    def __init__(self, chat_id, user_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'user_id': user_id
        }


class SetChatStickerSet(TelegramEndpoint):
    def __init__(self, chat_id, sticker_set_name):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'sticker_set_name': sticker_set_name
        }


class DeleteChatStickerSet(TelegramEndpoint):
    def __init__(self, chat_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id
        }



from abc import ABC


class TelegramEndpoint(ABC):
    method = 'GET'  # type: str
    endpoint = None  # type: str
    args = {}  # type: dict
    fields = {}  # type: dict

    def __init__(self):
        _classname = self.__class__.__name__
        self.endpoint = '{}{}'.format(_classname[0].lower, _classname[1:])

    def serialize(self):
        return self.method, self.endpoint, self.args, self.fields


from .chat import ExportChatInviteLink, SetChatPhoto, DeleteChatPhoto, SetChatTitle, SetChatDescription, PinChatMessage, UnpinChatMessage, LeaveChat, GetChat, GetChatAdministrators, GetChatMemberCount, GetChatMember, SetChatStickerSet, DeleteChatStickerSet  # noqa: F401, I101, E501
from .chat_action import SendChatAction  # noqa: F401, I101, E501
from .chat_member import KickChatMember, UnbanChatMember, RestrictChatMember, PromoteChatMember  # noqa: F401, I101, E501
from .contact import SendContact  # noqa: F401, I101, E501
from .file import EditMessageCaption, EditMessageMedia, GetFile, SendAudio, SendAnimation, SendDocument, SendMediaGroup, SendPhoto, SendVideo, SendVideoNote, SendVoice  # noqa: F401, I101, E501
from .forward import ForwardMessage  # noqa: F401, I101, E501
from .inline_mode import AnswerInlineQuery  # noqa: F401, I101, E501
from .location import SendLocation, EditMessageLiveLocation, StopMessageLiveLocation  # noqa: F401, I101, E501
from .reply_markup import InlineKeyboard, ReplyKeyboard, ReplyKeyboardRemove, ForceReply, AnswerCallbackQuery  # noqa: F401, I101, E501
from .text import SendMessage, EditMessageText  # noqa: F401, I101, E501
from .uncategorized import GetMe, GetUserProfilePhotos, EditMessageReplyMarkup, DeleteMessage  # noqa: F401, I101, E501
from .venue import SendVenue  # noqa: F401, I101, E501

from .exceptions import *  # noqa: F401, I101, F403, I100, I202



from . import TelegramEndpoint
from .commons import _ChatOrInlineMessage
from .reply_markup import ReplyMarkup


class GetMe(TelegramEndpoint):
    pass


class GetUserProfilePhotos(TelegramEndpoint):
    def __init__(self, user_id):
        super().__init__()
        self.args = {
            'user_id': user_id
        }

    def offset(self, offset):
        self.args = {
            **self.args,
            'offset': offset
        }
        return self

    def limit(self, limit):
        self.args = {
            **self.args,
            'limit': limit
        }
        return self


class EditMessageReplyMarkup(TelegramEndpoint, _ChatOrInlineMessage, ReplyMarkup):
    def __init__(self, chat_or_inline_message_id, message_id):
        super().__init__()
        self.__chat_or_inline_message(chat_or_inline_message_id, message_id)


class DeleteMessage(TelegramEndpoint):
    def __init__(self, chat_id, message_id):
        super().__init__()
        self.args = {
            'chat_id': chat_id,
            'message_id': message_id
        }



from .exceptions import InvalidParseMode


class _ChatOrInlineMessage(object):
    def __chat_or_inline_message(self, chat_or_inline_message_id, message_id):
        if message_id is not None:
            self.args = {
                **self.args,
                'chat_id': chat_or_inline_message_id,
                'message_id': message_id
            }
        else:
            self.args = {
                **self.args,
                'inline_message_id': chat_or_inline_message_id
            }


class ReplyToMessageId(object):
    # todo
    def reply_to_message_id(self, message_id):
        self.args = {
            **self.args,
            'reply_to_message_id': message_id
        }
        return self


class DisableNotification(object):
    def disable_notification(self):
        self.args = {
            **self.args,
            'disable_notification': True
        }
        return self


class ParseMode(object):
    def parse_mode(self, _parse_mode):
        if _parse_mode not in ['Markdown', 'HTML']:
            raise InvalidParseMode

        self.args = {
            **self.args,
            'parse_mode': _parse_mode
        }
        return self
